'''
# ------------------------------------------------------------------------

 Filter to create a new image simulating a CROCHET LOOK of the INPUT LAYER
 
 ......................................................................... 
 
 Release 1.0   initial
 Release 1.1   added BG transparent
               added work on selection
               added light ecru colour on crochet thread
 
 Flow implemented to get the final result:      
 
 1. save context values (restored at the end)
 2. create the output image (size is a user option, default=2400),
    BG as a copy of the input layer
    underBG for when working on selection
 3. create a working image and quantize it 
    thru changing mode to Indexed (nr of areas), then back to RGB
 4. loop on each area to create Border Inward Lines
    (using an embedded ancillary function derived from RobA "Mosaic Tile Helper")
 5. loop on each area to create Mosaic Tiles using Mosaic filter
 6. on the full image run Edge after a selective gaussian blur
    then invert values
 7. merge the three created textures to simulate a "Crochet look"
 8. create the Crochet Layer on the output image,
    by copying the outcome from the working image processing
 9. complete the final outcome according to the chosen type:
    - (0 = white crochet) : crochet on top of a BG filled
           with the average colour of the image, combined with a texture
    - (1 = nuanced crochet) : crochet colorized using the source layer
    - (2 = crochet on source) : crochet merged with the source on the BG
    - (3 = source on crochet) : source on top merged with the crochet
    - (4 = combo crochet) : options 0 and 2 combined
    on each one of the option take care of the options:
    - BG transparent
    - work on selection
    - thread colour
10. restore context
 
 ------------------------------------------------------------------
 COPYRIGHT NOTICE
 ------------------------------------------------------------------
 
 This program is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 3 of the License, or
 (at your option) any later version.  
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program if not, you can view the GNU General Public
 License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
 Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
 Ave, Cambridge, MA 02139, USA.
 
 --------------------------------------------------------------------------------------------------------------------------------------------------
 other info
 --------------------------------------------------------------------------------------------------------------------------------------------------

 --------------------------------------------------------------------------------------------------------------------------------------------------
 the following line terminates the long initial comments
'''

#!/usr/bin/env python
from gimpfu import *
import decimal

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# ========================================================
#   ANCILLARY MODULE : BORDER INWARD REPEAT (from RobA)
# --------------------------------------------------------
def BordersInwardRepeat (img, inLayer, inSpacing, inBSize, inColour) :
    
    width = img.width 
    height = img.height
    
    colourtoggle = 1          
     
    initial_selection = 0
    
    img.undo_group_start() 

    # save selection or select all if no selection
    if pdb.gimp_selection_is_empty (img) == False:
        initial_selection = pdb.gimp_selection_save (img)
    else:
        pdb.gimp_selection_all (img) 

    scratch_layer = inLayer.copy()
    img.add_layer (scratch_layer, -1)
    # fill the scratch layer with white
    scratch_layer.fill (WHITE_FILL)
    
    # work in scratch layer
    while pdb.gimp_selection_is_empty (img) == False:
        if colourtoggle == 0:
            gimp.set_foreground ((255, 255, 255, 255))
            colourtoggle = 1
        else:
            gimp.set_foreground ((0, 0, 0, 255))
            colourtoggle = 0
        pdb.gimp_edit_fill (scratch_layer, FOREGROUND_FILL)     
        pdb.gimp_selection_shrink (img, inSpacing)

    # select black
    pdb.gimp_image_select_color (img, CHANNEL_OP_REPLACE, scratch_layer, ((0, 0, 0, 255)))
    
    # turn it into a path
    
    pdb.plug_in_sel2path (img, scratch_layer)
    paths = img.vectors
    newpath = paths[0] 

    # make it visible and active
    newpath.visible = True
    pdb.gimp_image_set_active_vectors (img, newpath)
    # delete scratch layer
    img.remove_layer (scratch_layer)
    
    # load initial selection back up and delete the channel
    if initial_selection <> 0:
        pdb.gimp_selection_load (initial_selection)
        img.remove_channel (initial_selection)
    else:
        pdb.gimp_selection_none (img) # else clear the selection

    gimp.set_foreground (inColour)
    pdb.gimp_context_set_brush ("2. Hardness 100")
    pdb.gimp_context_set_brush_angle (0.00)
    pdb.gimp_context_set_brush_aspect_ratio (0.00)
    pdb.gimp_context_set_brush_size (inBSize)
    pdb.gimp_context_set_opacity (100.0)
    pdb.gimp_context_set_dynamics ("Dynamics Off")
    
    pdb.gimp_edit_stroke_vectors (inLayer, newpath)

    pdb.gimp_image_remove_vectors (img, newpath)
 
    #done =======================================================

    img.undo_group_end()
    
    return (inLayer)

############################################################################

def CrazyCrochet (inImage, inLayer,
                  widthheight,
                  nrAreas,
                  which_outcome,
                  bg, threadc) :
 
    pdb.gimp_context_push
    currFG = pdb.gimp_context_get_foreground()
    currBG = pdb.gimp_context_get_background()
    
    # set sample context
    pdb.gimp_context_set_sample_criterion (SELECT_CRITERION_COMPOSITE)
    pdb.gimp_context_set_sample_merged (False)
    pdb.gimp_context_set_sample_threshold (0.15)
    pdb.gimp_context_set_sample_threshold_int (15)
    pdb.gimp_context_set_sample_transparent (False)
     
    owidth = inImage.width
    oheight = inImage.height
    type = inImage.base_type
    name = inLayer.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    suffix = ", as a Crazy Crochet (colour areas ="+str(int(nrAreas))+")"
    
    # calculate scale factor for the new output image
    realWidthHeight = owidth+oheight      # real size of the input image
    if widthheight == 0 :   # means the user wants to have same size as input
        widthheight = realWidthHeight                                    
        scale = 1
    else :     
        optiWidthHeight = widthheight       # size to work on to get best results
        scale = decimal.Decimal(optiWidthHeight)/decimal.Decimal(realWidthHeight)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    # note : original will be scaled back after copy 
    pdb.gimp_image_scale (inImage, rwidth, rheight)
    
    if pdb.gimp_selection_is_empty (inImage) == True:
        workingon = "fulldrawable"
    else :
        workingon = "selection"
    
    (thereis, selx1, sely1, selx2, sely2) = pdb.gimp_selection_bounds (inImage)
     
# --------------------------------------------------------------------------------------

    newImage = pdb.gimp_image_new (rwidth, rheight, type)                    
    newBG = pdb.gimp_layer_new (newImage, rwidth, rheight, RGBA_IMAGE, "BG", 100.0, NORMAL_MODE)     # background layer
    newImage.add_layer (newBG, 0)
    
    underBG = pdb.gimp_layer_new_from_drawable (inLayer, newImage)
    newImage.add_layer (underBG, 1)                                   # to keep the non selected area of the source layer below
    underBG.name = "Under BG"
    alpha=pdb.gimp_drawable_has_alpha(underBG)
    if (alpha==False):
        pdb.gimp_layer_add_alpha (underBG)                            #  RGBA
    '''
    # find a colour which (hopefully) does not exist on the image, to use for filling the newBG
    pdb.gimp_context_set_sample_threshold (0.01)
    Rf = 11; Gf = 22; Bf = 33
    fcolour = (Rf, Gf, Bf, 255)
    pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE, underBG, (fcolour))
    isempty = pdb.gimp_selection_is_empty (newImage)
    while (Bf<160) and (isempty==False) :
        Rf += 1; Gf += 1; Bf += 1
        fcolour = (Rf, Gf, Bf, 255)
        pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE, underBG, (fcolour))
        isempty = pdb.gimp_selection_is_empty (newImage)
    #info_msg (str(fcolour))
    pdb.gimp_context_set_sample_threshold (0.15)    # restore
    '''
    # the above does not work ?!
    fcolour = (255, 0, 0, 255)
    pdb.gimp_context_set_background (fcolour)
    pdb.gimp_drawable_fill (newBG, BACKGROUND_FILL)
    
    pdb.gimp_selection_none (newImage)
    pdb.gimp_edit_copy (inLayer)
    floatsel = pdb.gimp_edit_paste (newBG, True)
    pdb.gimp_floating_sel_anchor (floatsel)
    
    #pdb.gimp_display_new (newImage)
    #pdb.gimp_displays_flush
    #return
   
# calculation for translating --------------
    selwidth = selx2-selx1
    selheight = sely2-sely1
    selx1n = int((rwidth-selwidth)/2)
    sely1n = int((rheight-selheight)/2)
    selx2n = selx1n+selwidth
    sely2n = sely1n+selheight
# end calculation --------------------------
 
    deltax = selx1-selx1n
    deltay = sely1-sely1n
    
    deltax2 = selx2-selx2n
    deltay2 = sely2-sely2n
    '''
    info_msg ("selx1, selx1n = "+str(selx1)+","+str(selx1n))
    info_msg ("sely1, sely1n = "+str(sely1)+","+str(sely1n))
    info_msg ("selx2, selx2n = "+str(selx2)+","+str(selx2n))
    info_msg ("sely2, sely2n = "+str(sely2)+","+str(sely2n))
    info_msg ("deltax, deltay = "+str(deltax)+","+str(deltay))
    info_msg ("deltax2, deltay2 = "+str(deltax2)+","+str(deltay2))
    '''
    #pdb.gimp_display_new (newImage)
    #pdb.gimp_displays_flush
    #return
  
# new selection for the translate ONLY
    pdb.gimp_image_select_rectangle (newImage, CHANNEL_OP_REPLACE, selx1n, sely1n, selwidth, selheight)
    
    pdb.gimp_selection_translate (newImage, deltax, deltay)
    pdb.gimp_drawable_offset (newBG, 0, 0, deltax, deltay)
    
    #pdb.gimp_display_new (newImage)
    #pdb.gimp_displays_flush
    #return
    
    newBG.mode = NORMAL_MODE
    newBG.name = "Background Image Layer"
    newBG.opacity = 100.0 
     
    alpha=pdb.gimp_drawable_has_alpha(newBG)
    if (alpha==False):
        pdb.gimp_layer_add_alpha (newBG)                            # RGBA
    pdb.gimp_image_undo_freeze (newImage)                           # save time and space
    
    pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE, newBG, (fcolour))
    pdb.gimp_selection_invert (newImage)
    
    if workingon == "selection" :
        pdb.gimp_edit_cut (underBG)                                 # keep ONLY the non selected area of the source layer 
    # save selection
    saved_selection = pdb.gimp_selection_save (newImage)
    pdb.gimp_selection_none (newImage)
    pdb.gimp_image_set_active_layer (newImage, newBG)
    
    #pdb.gimp_display_new (newImage)
    #pdb.gimp_displays_flush
    #return
    
#####################################################################################################
#   quantize the colour_areas using mode indexed on a duplicated image (WORK)
# =================================================================================================== 

    WORK_IMAGE = pdb.gimp_image_duplicate (newImage)
    WORK_DRAWABLE = WORK_IMAGE.active_layer
    WORK_DRAWABLE.name = "BG Layer"
    
    # change image mode to indexed (this produces a cartoonized image, nrAreas colors)
    
    pdb.gimp_image_convert_indexed (WORK_IMAGE, 0, 0, nrAreas, 0, 0, "ignored")    # INPUT   <<<<<<<< 
    is_indexed = pdb.gimp_drawable_is_indexed (WORK_DRAWABLE) 
    if is_indexed == 1:
        hidden_message = "OK, converted"
    else:    
        erro_msg ("ERROR! image not converted, exit")
        return
    #   
    # get the number of pixels in the colormap (which exists for indexed images)
    # ==========================================================================
    (count, colors) = pdb.gimp_image_get_colormap (WORK_IMAGE)
    # the count is nr_of_colors * 3, example 8*3=24
    nr_areas = (count/3) 
    #info_msg ("nr of colors in the palette: "+str(nr_areas)) 
  
    areas_palette = pdb.gimp_palette_new(pdb.gimp_image_get_name (WORK_IMAGE))
    
    index = 0
    while index < count:
    # Compute the color
        color = (colors[index], colors[index+1], colors[index+2])
    # Add to the filter temporary palette
        pdb.gimp_palette_add_entry (areas_palette, str(index/3), color)
        index+=3
        
    # Endwhile index<count
    # the temporary palette has been created            
    # ======================================

    # re-set image to RGB to proceed
    # ==============================
    pdb.gimp_image_convert_rgb (WORK_IMAGE)
    alpha=pdb.gimp_drawable_has_alpha (WORK_DRAWABLE) 
    if alpha == 0:
        pdb.gimp_layer_add_alpha (WORK_DRAWABLE)                        #  RGBA

    # set-up parameters on the base of the dimension
    # ----------------------------------------------
    # best values for recommended size 2400:
    #   minspace = 12, avgspace = 18, maxspace = 24
    #   mintile = 12, avgtile = 18, maxtile = 24
    #   bsize = 1.5 (*2), theight = 1.5, tspace = 1.5
    # hence these ratios apply:
    #   minspace = size*12/2400, avgspace = size*18/2400, maxspace = size*24/2400
    #   mintile = minspace, avgtile = avgspace, maxtile = maxspace
    #   bsize = size*15/24000, theight = bsize, tspace = bsize, spread = bsize
    #
    # five different crochet squares
    minspace = round (widthheight*12.0/2400.0,1)
    av1space = round (widthheight*15.0/2400.0,1)
    av2space = round (widthheight*18.0/2400.0,1)     
    av3space = round (widthheight*21.0/2400.0,1)
    maxspace = round (widthheight*24.0/2400.0,1)     
    
    mintile = int(minspace); av1tile = int(av1space); av2tile = int(av2space); av3tile = int(av3space); maxtile = int(maxspace)
    bsize = round (widthheight*15.0/24000.0, 2)
    theight = bsize
    tspace = bsize
    spread = bsize
    #info_msg ("minspace, av1space, av2space, av3space, maxspace = "+
    #           str(minspace)+","+str(av1space)+","+str(av2space)+","+str(av3space)+","+str(maxspace))
    #info_msg ("bsize = "+str(bsize))

#   =========================================
#   loop to get the meanvalues of the N areas                 
#   =========================================

    areaNR = 0
    minmean = 255
    maxmean = 0
    summean = 0
    areas = 0
    means = []
    for areaNR in range (0, nr_areas):
        one_color = pdb.gimp_palette_entry_get_color (areas_palette, areaNR)
        pdb.gimp_image_select_color (WORK_IMAGE, CHANNEL_OP_REPLACE, WORK_DRAWABLE, one_color)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds (WORK_IMAGE)
        if there_is_a_selection == 1:
            (mean, stddev, median, pixels, count, percent) = pdb.gimp_histogram (WORK_DRAWABLE, 0, 0, 255)
            means.append (int(round(mean)))
            summean += int(round(mean))
            if int(round(mean)) < minmean:
                minmean = int(round(mean))
            if int(round(mean)) > maxmean:
                maxmean = int(round(mean))
            areas += 1
    #info_msg ("end loop")
    meansasc = sorted(means, key=int)               # ascending order
    if areas <= 5:
        stop0 = meansasc[0]
        stop1 = meansasc[1]
        stop2 = meansasc[2]
        if areas == 3:
            stop3 = 255
            stop4 = 255
        if areas == 4:
            stop3 = meansasc[3]
            stop4 = 255
        elif areas == 5:
            stop3 = meansasc[3]
            stop4 = meansasc[4]
    else:
        stepm = int(float(maxmean-minmean)/5.0)
        stop1 = minmean+stepm
        stop2 = stop1+stepm
        stop3 = stop2+stepm
        stop4 = stop3+stepm
        
    #info_msg ("points = "+str(minmean)+","+str(stop1)+","+str(stop2)+","+str(stop3)+","+str(stop4)+","+str(maxmean))

#   ==================================    
#   loop to DRAW the N areas for tiles
#   part 1: borders inward
#   ================================== 

    pdb.gimp_image_undo_freeze (WORK_IMAGE)                      # save time and space
    
    WORK_TILES = WORK_DRAWABLE.copy()
    WORK_IMAGE.add_layer (WORK_TILES, 0)
    WORK_TILES.name = "Layer for Tiles (RobA)"

    areaNR = 0
    area_color = []   # empty list
    for areaNR in range (0, nr_areas):
        one_color = pdb.gimp_palette_entry_get_color (areas_palette, areaNR)
        area_color.append (one_color)
        (Ra,Ga,Ba,Aa) = one_color 

        pdb.gimp_image_select_color (WORK_IMAGE, CHANNEL_OP_REPLACE, WORK_TILES, one_color)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds (WORK_IMAGE)
        if there_is_a_selection == 1:
             
            color_area = pdb.gimp_image_get_selection (WORK_IMAGE)
            this_channel = pdb.gimp_selection_save (WORK_IMAGE)
             
            (mean, stddev, median, pixels, count, percent) = pdb.gimp_histogram (WORK_TILES, 0, 0, 255)
            if int(round(mean)) >= stop4 :
                spacing = minspace # light areas
            elif int(round(mean)) >= stop3 :
                spacing = av1space # medium-light areas
            elif int(round(mean)) >= stop2 :
                spacing = av2space # medium areas
            elif int(round(mean)) >= stop1 :
                spacing = av3space # medium-dark areas
            else :
                spacing = maxspace # dark areas
                
            #info_msg ("area nr "+str(areaNR)+" mean = "+str(int(round(mean)))+" spacing = "+str(spacing))   
            pdb.gimp_context_set_brush ("2. Hardness 100")
            pdb.gimp_context_set_brush_size (bsize)       

            BordersInwardRepeat (WORK_IMAGE, WORK_TILES, spacing, bsize*1.5, ((0,0,0,255)))
             
    #info_msg ("end loop")
    pdb.gimp_selection_none (WORK_IMAGE)
    pdb.plug_in_spread (WORK_IMAGE, WORK_TILES, spread, spread)
    
    #pdb.gimp_display_new (WORK_IMAGE)
    #return
    
#   ==================================    
#   loop to DRAW the N areas for tiles
#   part 2: mosaic tiles
#   ==================================

    WORK_MOSAIC = WORK_DRAWABLE.copy()
    WORK_IMAGE.add_layer (WORK_MOSAIC, 0)
    pdb.gimp_layer_flatten (WORK_MOSAIC)    # remove ALPHA for MOSAIC
    WORK_MOSAIC.name = "Layer for Mosaic"

    areaNR = 0
    area_color = []   # empty list
    for areaNR in range (0, nr_areas):
        one_color = pdb.gimp_palette_entry_get_color (areas_palette, areaNR)
        area_color.append (one_color)
        (Ra,Ga,Ba,Aa) = one_color 
        pdb.gimp_image_select_color (WORK_IMAGE, CHANNEL_OP_REPLACE, WORK_MOSAIC, one_color)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds (WORK_IMAGE)
        if there_is_a_selection == 1:
            
            color_area = pdb.gimp_image_get_selection (WORK_IMAGE)
            this_channel = pdb.gimp_selection_save (WORK_IMAGE)
            
            pdb.gimp_context_set_brush ("2. Hardness 100")
            pdb.gimp_context_set_brush_size (bsize)        

            # adaptive parameters
            (mean, stddev, median, pixels, count, percent) = pdb.gimp_histogram (WORK_MOSAIC, 0, 0, 255)
            if int(round(mean)) >= stop4 :
                tilesize = mintile # light areas
            elif int(round(mean)) >= stop3 :
                tilesize = av1tile # medium light areas
            elif int(round(mean)) >= stop2 :
                tilesize = av2tile # medium areas
            elif int(round(mean)) >= stop1 :
                tilesize = av3tile # medium dark areas    
            else:
                tilesize = maxtile # dark areas
            
            #info_msg ("area nr "+str(areaNR)+" mean = "+str(int(round(mean)))+" tilesize = "+str(tilesize))   
    
            height = theight # fixed
            space = tspace # fixed
            neat = 0.45
            #info_msg ("tilesize,height,space,neat="+str(tilesize)+","+str(height)+","+str(space)+","+str(neat))
            pdb.plug_in_mosaic (WORK_IMAGE, WORK_MOSAIC,
                        tilesize, height, space, neat,    # tile characteristics: size, height, space, neat      # was 4,1.0.25
                        True,                        # tile-split
                        90.0,                       # light
                        0.0,                         # colour variation     # was 0.4
                        False, True,                 # antialias, colour averaging      # was True,False
                        2, 0, 0)                     # octagons, surface, grout-colour (0=BG, 1=FG/BG)
            
    #info_msg ("end loop")
    pdb.gimp_palette_delete(areas_palette)
    pdb.gimp_selection_none (WORK_IMAGE)
    pdb.plug_in_spread (WORK_IMAGE, WORK_MOSAIC, spread, spread)
 
#   ================================   
#   DRAW the countours of the areas
#   ================================ 
     
    WORK_CONTOURS = WORK_DRAWABLE.copy()
    WORK_IMAGE.add_layer (WORK_CONTOURS, 0)
    pdb.gimp_layer_flatten (WORK_CONTOURS)     
    WORK_CONTOURS.name = "Layer for Contours"
    pdb.plug_in_sel_gauss (WORK_IMAGE, WORK_CONTOURS, 8.0, 40) 
    pdb.plug_in_edge (WORK_IMAGE, WORK_CONTOURS, 10.0, 0, 0)
    pdb.gimp_threshold (WORK_CONTOURS, 127, 255)
    pdb.plug_in_vinvert (WORK_IMAGE, WORK_CONTOURS)
    pdb.plug_in_erode (WORK_IMAGE, WORK_CONTOURS, 1, HISTOGRAM_VALUE, 1.0, 7, 0, 128)

#   ============================= 
#   merge to create the "texture"
#   =============================

    WORK_CONTOURS.opacity = 33.3
    WORK_CONTOURS.mode = GRAIN_MERGE_MODE
    
    WORK_MOSAIC.opacity = 100.0  
    WORK_MOSAIC.mode = DARKEN_ONLY_MODE
    
    pdb.gimp_item_set_visible (WORK_DRAWABLE, False)
    
    #pdb.gimp_display_new (WORK_IMAGE)
    #return
    
    WORK_MERGED = pdb.gimp_image_merge_down (WORK_IMAGE, WORK_MOSAIC, CLIP_TO_BOTTOM_LAYER)
    WORK_MERGED.name = "Merged Layers Mosaic and Tiles"
    
    WORK_MERGED = pdb.gimp_image_merge_down (WORK_IMAGE, WORK_CONTOURS, CLIP_TO_BOTTOM_LAYER)
    WORK_MERGED.name = "Merged Layers Contours, Mosaic and Tiles"
    
    pdb.plug_in_emboss (WORK_IMAGE, WORK_MERGED, 60.0, 45.0, 20.0, 1)
    pdb.plug_in_vinvert (WORK_IMAGE, WORK_MERGED)
    
    #pdb.gimp_display_new (WORK_IMAGE)
    #return
    
#================================================================================================ 
#   COPY THE OUTCOME OF THE PROCESSED WORKING IMAGE TO THE OUTPUT IMAGE    
# =============================================================================================== 
    
    CrochetLayer = pdb.gimp_layer_new_from_drawable (WORK_MERGED, newImage)
    pdb.gimp_image_delete (WORK_IMAGE)          # <<<<<<<<<<<<<<<<<<<<<<<<< not needed any more
    CrochetLayer.name =  "Crochet Layer"
    newImage.add_layer (CrochetLayer, 0)
    CrochetLayer.mode =  NORMAL_MODE
    
    # ===================
    # which final outcome
    # -------------------
    
    pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, saved_selection)
    
    if workingon == "selection" :
        pdb.gimp_selection_invert (newImage)
        pdb.gimp_edit_cut (CrochetLayer)
        pdb.gimp_edit_cut (newBG)
        pdb.gimp_selection_invert (newImage)
    
    pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE, CrochetLayer, ((255, 255, 255, 255)))
    only_crochet_selection = pdb.gimp_selection_save (newImage)         # to restore later
    
    if threadc == 1:    # unbleached white
        ecru = (254, 244, 228, 255)     # light ecru
        pdb.gimp_context_set_foreground (ecru)
        pdb.gimp_edit_fill (CrochetLayer, FOREGROUND_FILL)
        
    pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, saved_selection)
    
    #pdb.gimp_display_new (newImage)
    #pdb.gimp_displays_flush
    #return
    
    if which_outcome == 0 :     # white crochet
        if bg == 0: # auto-generated
            pdb.gimp_context_set_foreground ((00,32,00,255))
            (meanR, stddev, median, pixels, count, percent) = pdb.gimp_histogram (newBG, 1, 0, 255)
            (meanG, stddev, median, pixels, count, percent) = pdb.gimp_histogram (newBG, 2, 0, 255)
            (meanB, stddev, median, pixels, count, percent) = pdb.gimp_histogram (newBG, 3, 0, 255)
            BGcolour = (max(int(meanR-100),0),max(int(meanG-100),0),max(int(meanB-100),0),255)
            #info_msg ("colour = "+str(BGcolour))
            pdb.gimp_context_set_foreground (BGcolour)
            pdb.gimp_edit_fill (newBG, FOREGROUND_FILL)
            pdb.gimp_context_set_pattern ("3D Green")
            newBG2 = pdb.gimp_layer_new (newImage, rwidth, rheight, RGB_IMAGE, "pattern BG", 100.0, NORMAL_MODE)
            newImage.add_layer (newBG2, 1)
            pdb.gimp_edit_fill (newBG2, PATTERN_FILL)
            #pdb.gimp_desaturate (newBG2)
            pdb.gimp_brightness_contrast (newBG2, +50, +75)
            newBG2.mode = GRAIN_MERGE_MODE
            newBG2.opacity = 100.00
            newBG = pdb.gimp_image_merge_down (newImage, newBG2, CLIP_TO_BOTTOM_LAYER)
            pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, saved_selection)
            pdb.gimp_selection_invert (newImage)
            if workingon == "selection" :
                pdb.gimp_edit_cut (newBG)
            CrochetLayer.mode = HARDLIGHT_MODE
            
        if bg==1:  # transparent
            CrochetLayer.mode = NORMAL_MODE
            pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, only_crochet_selection)
            pdb.gimp_selection_invert (newImage)
            pdb.gimp_selection_shrink (newImage, 1)            
            pdb.gimp_edit_cut (CrochetLayer)
            pdb.gimp_edit_cut (newBG)
            pdb.gimp_edit_cut (underBG)
        # end transparent
        #pdb.gimp_display_new (newImage)
        #pdb.gimp_displays_flush
        #return
        FinalL = pdb.gimp_image_merge_down (newImage, newBG, CLIP_TO_BOTTOM_LAYER)
        FinalL = pdb.gimp_image_merge_down (newImage, CrochetLayer, CLIP_TO_BOTTOM_LAYER)
        suffix = suffix + " (white crochet)"
    if which_outcome == 1 :     # nuanced crochet
        pdb.gimp_image_raise_item_to_top (newImage, newBG)
        newBG.mode = HARDLIGHT_MODE
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, only_crochet_selection)
        pdb.gimp_selection_invert (newImage)
        pdb.gimp_selection_shrink (newImage, 1)
        pdb.gimp_edit_cut (newBG)
        pdb.gimp_threshold (CrochetLayer, 128, 255)
        #pdb.gimp_display_new (newImage)
        #pdb.gimp_displays_flush
        #return
        MergedL = pdb.gimp_image_merge_down (newImage, newBG, CLIP_TO_BOTTOM_LAYER)
        VisibleL = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible")
        newImage.add_layer (VisibleL, 0)
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, saved_selection)
        pdb.gimp_selection_invert (newImage)
        if workingon == "selection" :
            pdb.gimp_edit_cut (VisibleL)
        pdb.gimp_selection_none (newImage)
        VisibleL.mode = HARDLIGHT_MODE
        VisibleL.opacity = 75.00
        if bg==1:  # transparent
            pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, only_crochet_selection)
            pdb.gimp_selection_invert (newImage)
            pdb.gimp_selection_shrink (newImage, 1)            
            pdb.gimp_edit_cut (VisibleL)
            pdb.gimp_edit_cut (MergedL)
            pdb.gimp_edit_cut (underBG)
        #pdb.gimp_display_new (newImage)
        #pdb.gimp_displays_flush
        #return
        FinalL = pdb.gimp_image_merge_down (newImage, MergedL, CLIP_TO_BOTTOM_LAYER)
        FinalL = pdb.gimp_image_merge_down (newImage, VisibleL, CLIP_TO_BOTTOM_LAYER)
        suffix = suffix + " (nuanced crochet)"
    if which_outcome == 2 :     # crochet on source
        CrochetLayer.mode = GRAIN_MERGE_MODE
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, saved_selection)
        pdb.gimp_selection_invert (newImage)
        if workingon == "selection" :
            pdb.gimp_edit_cut (newBG)
        pdb.gimp_hue_saturation (newBG, 0, 0, -35, +70)
        MergedL = pdb.gimp_image_merge_down (newImage, CrochetLayer, CLIP_TO_BOTTOM_LAYER)
        if bg==1:  # transparent
            pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, only_crochet_selection)
            pdb.gimp_selection_invert (newImage)
            pdb.gimp_selection_shrink (newImage, 1)            
            pdb.gimp_edit_cut (MergedL)
            pdb.gimp_edit_cut (underBG)
        FinalL = pdb.gimp_image_merge_down (newImage, MergedL, CLIP_TO_BOTTOM_LAYER)
        suffix = suffix + " (crochet on source)"
    if which_outcome == 3 :     # source on crochet
        pdb.gimp_image_raise_item_to_top (newImage, newBG)
        pdb.gimp_hue_saturation (newBG, 0, 0, -25, +25)
        newBG.mode = HARDLIGHT_MODE
        MergedL = pdb.gimp_image_merge_down (newImage, newBG, CLIP_TO_BOTTOM_LAYER)
        if bg==1:  # transparent
            pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, only_crochet_selection)
            pdb.gimp_selection_invert (newImage)
            pdb.gimp_selection_shrink (newImage, 1)            
            pdb.gimp_edit_cut (MergedL)
            pdb.gimp_edit_cut (underBG)
        FinalL = pdb.gimp_image_merge_down (newImage, MergedL, CLIP_TO_BOTTOM_LAYER)
        suffix = suffix + " (source on crochet)"
    if which_outcome == 4 :     # COMBO: crochet on source + white crochet+
        CrochetLayer.mode = GRAIN_MERGE_MODE
        DupCrochetLayer = CrochetLayer.copy()
        newImage.add_layer (DupCrochetLayer, 0)
        CosL = pdb.gimp_image_merge_down (newImage, CrochetLayer, CLIP_TO_BOTTOM_LAYER)
        DupCrochetLayer.mode = BURN_MODE
        MergedL = pdb.gimp_image_merge_down (newImage, DupCrochetLayer, CLIP_TO_BOTTOM_LAYER)
        if bg==1:  # transparent
            pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, only_crochet_selection)
            pdb.gimp_selection_invert (newImage)
            pdb.gimp_selection_shrink (newImage, 1)            
            pdb.gimp_edit_cut (MergedL)
            pdb.gimp_edit_cut (underBG)
        FinalL = pdb.gimp_image_merge_down (newImage, MergedL, CLIP_TO_BOTTOM_LAYER)
        suffix = suffix + " (combo crochet)"
    # ----------------------------------------------------------------------------------------------- 
    FinalL.name = name+suffix
    
    if threadc == 1 :     #unbleached, check whether there is a residual full white thread
        pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE, FinalL, ((255, 255, 255, 255)))
        if pdb.gimp_selection_is_empty (newImage) == False:  # there is a residual full white thread
            ecru = (254, 244, 228, 255)     # light ecru
            pdb.gimp_context_set_foreground (ecru)
            pdb.gimp_edit_fill (FinalL, FOREGROUND_FILL)

    pdb.gimp_selection_none (newImage)
    pdb.gimp_display_new (newImage)
    
    pdb.gimp_image_scale (inImage, owidth, oheight)                 # source back to original size
    if workingon == "fulldrawable" :
        pdb.gimp_selection_none (inImage)
    pdb.gimp_displays_flush

    pdb.gimp_context_pop
    pdb.gimp_context_set_foreground(currFG)
    pdb.gimp_context_set_background(currBG)
    
    #return()
    return (FinalL)

# =====================================================================================================

register(
    "CrazyCrochet",    
    "to simulate a crazy crochetted image (v.1.1)",   
    "This script transforms the input drawable into a crazy crochetted image, with or without colours (v.1.1).",
    "Diego", 
    "Diego Nassetti ", 
    "February 2017",
    "CrazyCrochet (v.1.1)", 
    "RGB*,GRAY*", 
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "idrawable", "Input drawable", None),
      (PF_ADJUSTMENT, "size", "Width+Height of New Image"+"\n (0=same as input)"+
                              "\n (>2000 recommended)", 2400, (0, 9600, 300)),
      (PF_ADJUSTMENT, "areas",  "Texture Areas", 5, (3, 15, 1)),
      (PF_OPTION, "color", "Final Outcome option", 0, ("White Crochet",
                                                       "Nuanced Crochet",
                                                       "Crochet on Source",
                                                       "Source on Crochet",
                                                       "Combo",
                                                      )),
      (PF_OPTION, "bg", "Background for Crochet", 0, ("auto-generated",
                                                      "transparent",
                                                     )),
      (PF_OPTION, "threadc", "Colour of Crochet thread", 0, ("white",
                                                            "unbleached",
                                                            )),
    ], 
    [
      (PF_DRAWABLE, "odrawable", "Output drawable", None),  
    ],
    CrazyCrochet,
    menu="<Image>/Diego/Artistic",
    )

main()