'''

 Plug-in Title : "PrintOnStones2.py"
 Gimp Menu     : "<Image>/Diego/Artistic" 
 Release       :  1.0    (initial, named PrintOnStones)
                  2.0    (added support for non-Windows users)
                         (added a new contour: TT-hardbevel - thanks to TinTran)
 Author        :  Diego Nassetti (dinasset@gimpchat.com)
 Licence       :  GNU General Public License
 
 System Environment used in development 
 --------------------------------------                
 Platform      : Windows 10 64-bits
 Graphic Apps  : Gimp 2.8.22
                 no G'mic  
 Language      : python as included in Gimp
                 
 Extrn Filters : none
                
 
 Summary
 ------- 
  Creates an image looking as a print on stone  
  -------------------------------------------- 
  
 Flow implemented to get the final result 
 ----------------------------------------
  check OS environment (Windows/Linux)
  - define slash format according to the OS to properly access the files
  
  convert image to RGB if GRAY (restore at the end)
  create a new image by copying the source image
  resize according to thesize chosen by the user
  remove blacks using a curve
  check luminance and adjust when < 160 
  save adjusted base layer in a buffer for later use
  
  if there is no user selection:
  - increase the canvas by 4%
  - select the source image as working selection
  if there is an existing selection:
  - shrink it by 1 pixel to avoid problens when whole image is selected
  save selection dimensions and centre
  
  copy and desaturate the BG layer
  copy the desaturated layer and apply Emboss filter (DARKEN_ONLY)
  create a new layer from visible and add a bit of Brightness/Contrast
  
  create a new layer and paste the buffer saved in the preliminary phase
  set mode to COLOR
  duplicate the source and posterize (32 colours) thru an ancillary function
  increase its saturation
  set mode to HARDLIGHT
  create a new layer from visible and autocrop it (visible 2)
       
  create a new layer for the pattern
  user can either:
  - select a single pattern
  - let the filter choose among a set identified by a sub-string
    (textures must be available otherwise you get an error)
    (4 options available: best luma, best colour, best luma&colour, random)    
  apply the requested transformation to the chosen pattern:
  - rescale the patterm to the layer size
  - make it seamless (thru an ancillary function)
  - make it seamless and rescale it to the layer size
  - do nothing (layer is filled with the pattern as is)
  
  duplicate the visible 2 layer
  reduce opacity  
  convert to grayscale and remove white (colortoalpha) 
    
  duplicate the pattern layer and posterize thru an ancillary function
  set this 2nd pattern layer mode to HARDLIGHT
  reduce opacity
  reduce saturation
  
  if the user entered the filter with a selection:
  - create a new layer for the "frame"
    - fill with a std Gimp gradient 
    - decrease brightness and increase contrast
    - cut the selection area (only the area around is left)
    - set mode to MULTIPLY
    
  create a new layer from visible (visible 3)
  apply a carve effect if chisel script is available
  apply an hardbevel effect (TT) if chosen
  otherwise create a simple border around
  
  if the user entered the filter without a selection:
  - create a new layer for the "frame" (under-layer)
    - fill with a std Gimp bricks pattern, adjusting the size 
  
  last step: flatten the image if requested by the user (default=yes)
  restore
  
  -------------------------------------------------------------------
  
  This program is free software  you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation  either version 3 of the License, or
  (at your option) any later version.  
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY  without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program  if not, you can view the GNU General Public
  License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
  Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
  Ave, Cambridge, MA 02139, USA.
 
   
 -------------------------------------------------------------------
 other info
 -------------------------------------------------------------------
 
'''
#!/usr/bin/env python
from gimpfu import *
import decimal
import os
from os.path import expanduser
from os import listdir
import math
import random
import time
import datetime
import platform

############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

##################################################################################
# SERVICE ROUTINES
##################################################################################


def load_patterns(names_filter):    # load all patterns
    
    np = 0  # nr of patterns loaded
    LPATT = []  # list for mixed patterns
    (np, LPATT) = pdb.gimp_patterns_get_list (names_filter)       
    
    tempImage = pdb.gimp_image_new(256, 256, RGB)
    tempLayer = pdb.gimp_layer_new(tempImage,256,256,RGB_IMAGE,"temp",100.0,NORMAL_MODE)
    tempImage.add_layer(tempLayer, 0)

    LLUMA = []  # list for LUMA value
    LCOLR = []  # list for PREVAILING COLOURS 
    
    if (np == 0):
        erro_msg("ERROR! Folder for patterns does not contain any valid pattern. Exit.")
        exit
    else :
        #info_msg ("info: nr of patterns = "+str(np))
        for i in range (np) :
            if LPATT[i] <> "":  # don't load null name patterns (fived)
                pdb.gimp_context_set_pattern (LPATT[i])
                pdb.gimp_drawable_fill (tempLayer, PATTERN_FILL)
    
                (meanR, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 1, 0, 255)
                (meanG, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 2, 0, 255)
                (meanB, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 3, 0, 255)
                aCformula=int(float(meanR*0.299)+float(meanG*0.587)+float(meanB*0.114))       #luminance=(red*0.299+green*0.587+blue*0.114)
                LLUMA.append(aCformula)
            
                meanCol = find_predominant_color (meanR, meanG, meanB)
                #info_msg("the predominant colour of this filler is "+meanCol)
                LCOLR.append(meanCol)        
            
    il=len(LCOLR)
    #for i in range(0, il):
        #info_msg(LCOLR[i])
    return(LPATT,LLUMA,LCOLR)

def find_predominant_color (R, G, B):
    
    R16=int(R/16)
    G16=int(G/16)
    B16=int(B/16)
    maxRGB=max(R16,G16,B16)
    minRGB=min(R16,G16,B16)
    limit = 0.25
    # normalize values if differences are low
    if (maxRGB-minRGB) < int(maxRGB*limit):  # if difference lower than limit of the maximum, decide it's GRAY
        search = "WHT"
    elif (maxRGB==R16):
        if ((R16-G16) >= int(R16*limit)) & ((R16-B16) >= int(R16*limit)): 
            search="RED" 
        elif (R16-B16) >= int(R16*limit):
            search="YEL"
        else:
            search="MAG"
    elif (maxRGB==G16):
        if ((G16-R16) >= int(G16*limit)) & ((G16-B16) >= int(G16*limit)): 
            search="GRE"
        elif (G16-B16) >= int(G16*limit):
            search="YEL"
        else:
            search="TRQ"
    elif (maxRGB==B16):
        if ((B16-R16) >= int(B16*limit)) & ((B16-G16) >= int(B16*limit)): 
            search="BLU"
        elif (B16-G16) >= int(B16*limit):
            search="MAG"
        else:
            search="TRQ"
    else:   # placeholder as default impossible case
        search="WHT"
    #info_msg("the predominant colour of this area is "+search)
    return(search)

    
def choose_appropriate_texture_by_LUMACOLOUR (areaColor, LCOLR, LLUMA, LTEXT):
            
    LLUMAS=[]   # subset of Luma
    LTEXTS=[]   # subset of Names
    ss=0        # default
    #info_msg("areaColor = "+str(areaColor))
    (R, G, B, T)=areaColor
    search = find_predominant_color (R, G, B)
    #info_msg("the predominant colour of this area is "+search)
    il=len(LCOLR)
    for i in range(0, il):
        if search==LCOLR[i]:
            chosenColr=LCOLR[i] 
            chosenLuma=LLUMA[i]
            chosenName=LTEXT[i]
            LLUMAS.append(chosenLuma)
            LTEXTS.append(chosenName)
            ss+=1
    
    if ss == 0:    # it may happen that there is no similar colour
        # use random texture
        rn = random.randrange (0, len(LTEXT))
        textureName=LTEXT[rn]
        info_msg ("warning: no appropriate colour found, chosen texture random = "+textureName)
    else:    
        textureName=choose_appropriate_texture_by_LUMA(areaColor, LLUMAS, LTEXTS)
    return(textureName)

def choose_appropriate_texture_by_LUMA (areaColor, LLUMA, LTEXT):
            
    closest="**NONE**"              # default
    distance=999999                 # limit
    (R, G, B, T)=areaColor 
    areaf=int(float(R*0.299)+float(G*0.587)+float(B*0.114))  #new, based on luminosity
    ##info_msg("this area color formula is "+str(areaf))
    il=len(LLUMA)
    for i in range(0, il):
        this_distance=abs(areaf-LLUMA[i])
        ##info_msg("the distance from "+LP[i]+" is "+str(this_distance))
        if this_distance<distance:
            distance=this_distance
            closest=LTEXT[i]
    textureName=closest
    #info_msg("closest = "+textureName)
    
    return(textureName)

def choose_appropriate_texture_by_COLOUR (areaColor, LCOLR, LTEXT):
            
    chosen="**NONE**"              # default
    #info_msg("areaColor = "+str(areaColor))
    (R, G, B, T)=areaColor
    search = find_predominant_color (R, G, B)
    #info_msg("the predominant colour of this area is "+search)
    il=len(LCOLR)
    for i in range(0, il):
        if search==LCOLR[i]:
            chosen=LCOLR[i] 
            textureName=LTEXT[i]
            #info_msg("chosen = "+textureName)
            break
    
    if chosen == "**NONE**":    # it may happen that there is no similar colour
        # use random texture
        rn = random.randrange (0, len(LTEXT))
        textureName=LTEXT[rn]
        #info_msg ("chosen random = "+textureName)
        
    return(textureName)

def posterizeIT (SourceLayer, nrofcol) :
    
    SourceImage = pdb.gimp_item_get_image (SourceLayer)
    width = SourceLayer.width
    height = SourceLayer.height
    wrkImage = pdb.gimp_image_new(width, height, RGB)
    wrkBG = pdb.gimp_layer_new_from_drawable(SourceLayer, wrkImage)     # background layer
    wrkBG.mode = NORMAL_MODE
    wrkBG.name = "Background Work Layer"
    wrkBG.opacity = 100.0 
    wrkImage.add_layer(wrkBG, 0)
    pdb.gimp_image_convert_indexed(wrkImage, 0, 0, nrofcol, 0, 0, "ignored")
    is_indexed=pdb.gimp_drawable_is_indexed(wrkBG) 
    if is_indexed==1:
        hidden_message="OK, converted"
    else:    
        pdb.gimp_message("ERROR! layer not converted, exit")
        return
# re_set image to RGB for returning the layer
    pdb.gimp_image_convert_rgb(wrkImage)
    SourceLayer2 = pdb.gimp_layer_new_from_drawable(wrkBG, SourceImage) 
    
    return (SourceLayer2)


def seamlessIT (Pattern, Slash) :       # added Slash for compatibility with Linux OS
    
    (patternWidth, patternHeight, bpp) = pdb.gimp_pattern_get_info (Pattern)
    temp_image = pdb.gimp_image_new (patternWidth, patternHeight, RGB) 
    temp_layer = pdb.gimp_layer_new (temp_image, patternWidth, patternHeight,
                                     RGBA_IMAGE, "Choosen texture", 100.00, NORMAL_MODE)
    pdb.gimp_context_set_pattern (Pattern)
    pdb.gimp_drawable_fill (temp_layer, PATTERN_FILL)
    pdb.gimp_image_add_layer (temp_image, temp_layer, 0)
    
    pdb.gimp_context_set_foreground ((0,0,0,255))
    pdb.gimp_context_set_background ((255,255,255,255))
    
    temp_layer2 = temp_layer.copy()
    pdb.gimp_layer_add_alpha (temp_layer2)    #  RGBA
    temp_image.add_layer (temp_layer2, 0)
    pdb.gimp_item_transform_flip_simple (temp_layer2, ORIENTATION_HORIZONTAL, 0, patternWidth/2)
    mask = pdb.gimp_layer_create_mask(temp_layer2, ADD_WHITE_MASK)
    pdb.gimp_layer_add_mask(temp_layer2,mask)
    pdb.gimp_edit_blend (mask, FG_BG_RGB_MODE, NORMAL_MODE, GRADIENT_LINEAR, 100.0, 0, REPEAT_NONE,
                         False, False, 0, 0, 1, 0, patternHeight/2, patternWidth, patternHeight/2)
    temp_layer = pdb.gimp_image_merge_down (temp_image, temp_layer2, 0)
        
    temp_layer2 = temp_layer.copy()
    pdb.gimp_layer_add_alpha (temp_layer2)    #  RGBA
    temp_image.add_layer (temp_layer2, 0)
    pdb.gimp_item_transform_flip_simple (temp_layer2, ORIENTATION_VERTICAL, 0, patternHeight/2)
    mask = pdb.gimp_layer_create_mask(temp_layer2, ADD_WHITE_MASK)
    pdb.gimp_layer_add_mask(temp_layer2,mask)
    pdb.gimp_edit_blend (mask, FG_BG_RGB_MODE, NORMAL_MODE, GRADIENT_LINEAR, 100.0, 0, REPEAT_NONE,
                         False, False, 0, 0, 1, patternWidth/2, 0, patternWidth/2, patternHeight)
    temp_layer = pdb.gimp_image_merge_down (temp_image, temp_layer2, 0)
    #pdb.gimp_display_new (temp_image)
    #return
    temp_layerfull = pdb.gimp_image_flatten (temp_image)
    #pdb.gimp_display_new (temp_image)
    #return
    d_dir = pdb.gimp_gimprc_query ("gimp_dir")
    SLP_name = "temp_pat"
    #SLP_name2 = d_dir+"\\patterns\\"+SLP_name+".pat"
    SLP_name2 = d_dir+Slash+"patterns"+Slash+SLP_name+".pat"        # changed for compatibility with Linux OS
    pdb.file_pat_save (temp_image, temp_layerfull, SLP_name2, "", SLP_name)
    pdb.gimp_patterns_refresh()
    
    return (SLP_name)    # return the name of the temporary pattern created, seamless


##################################################################################
# MAIN PROGRAM
##################################################################################

def PrintOnStones2 (inImage, inLayer,
                            widthheight,
                            namesFilter,
                            singlePattern,					 
                            PatternAction,
                            selMethod,
                            which3d,
                            doFlatten):
  
    #### PRELIMINARY STEPS +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    #  Calculate and create a new image (size chosen by the user)
    
    pdb.gimp_image_undo_group_start(inImage)
     
    # set up image characteristics
    width = inImage.width
    height = inImage.height
    origtype = inImage.base_type
    
    if origtype == GRAY :
        pdb.gimp_image_convert_rgb (inImage)
        type = RGB
    else :
        type = origtype
    
    sourcename = inLayer.name
    if '.' in sourcename:
        split = sourcename.split(".", 1)
        sourcename = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H-%M-%S')
    
    # calculate scale factor for the new output image
    realWidthHeight = width+height      # real size of the input image
    if widthheight == 0 :               # means the user wants to have same size as input
        widthheight = realWidthHeight                                    
        scale = 1
    else :     
        optiWidthHeight = widthheight       # size to work on to get best results
        scale = decimal.Decimal(optiWidthHeight)/decimal.Decimal(realWidthHeight)      # ratio for scaling
    #info_msg (str(scale))

    pdb.gimp_context_push
    currFG = pdb.gimp_context_get_foreground()
    currBG = pdb.gimp_context_get_background()
    
    home = expanduser("~")
    # additions for compatibility with Linux systems
    myplatform=platform.system()
    #pdb.gimp_message("I'm running "+myplatform)
    if myplatform=="Linux":
        myslash="/"
    if myplatform=="Windows":
        myslash="\\"
    
    # type of 3d chosen
    if which3d == 0:    # simple contour
        doContour = True
        doChisel = False
        doHardbump = False
    elif which3d == 1:    # use chisel script
        doContour = False
        doChisel = True
        doHardbump = False
    else:    # use hardbump (embedded from TinTran) script
        doContour = False
        doChisel = False
        doHardbump = True
    
    # create output image
    
    newwidth=int(float(width*scale))
    newheight=int(float(height*scale))
    
    newImage = pdb.gimp_image_duplicate (inImage)                    
    newBG = newImage.active_layer                            
    newBG.name = "Base Layer"
    pdb.gimp_image_scale (newImage, newwidth, newheight)
    
    # visibility off for all layers except the active one
    LA = newImage.layers
    for alayer in LA:
        if alayer <> newBG:
            pdb.gimp_layer_set_visible (alayer, False)
 
    alpha=pdb.gimp_drawable_has_alpha(newBG)
    if (alpha==False):
        pdb.gimp_layer_add_alpha (newBG)    #  RGBA
        
    # remove blacks
    pdb.gimp_curves_spline (newBG, 0, 6, (0, 16, 127, 127, 255, 255))
    
    # calculate luminance
    (meanR, stddev, median, pixels, count, percent)=pdb.gimp_histogram(newBG, 1, 0, 255)
    (meanG, stddev, median, pixels, count, percent)=pdb.gimp_histogram(newBG, 2, 0, 255)
    (meanB, stddev, median, pixels, count, percent)=pdb.gimp_histogram(newBG, 3, 0, 255)
    aCformula=int(float(meanR*0.299)+float(meanG*0.587)+float(meanB*0.114))
    #info_msg ("source image luminance = "+str(aCformula))
    if aCformula < 160:
        tempBG2 = newBG.copy()
        newImage.add_layer (tempBG2, 0)
        tempBG2.mode = SCREEN_MODE
        tempBG2.opacity = min (100.0, 160.0-aCformula)
        newBG = pdb.gimp_image_merge_down (newImage, tempBG2, 2)
    
    # save adjusted base layer in a buffer
    
    realBufferName = pdb.gimp_edit_named_copy (newBG, "myBufferName")
    
    pdb.gimp_context_set_feather (False)
    pdb.gimp_context_set_feather_radius (0.00, 0.00)
    
    noUsrSel = pdb.gimp_selection_is_empty (newImage)
    if noUsrSel == True:   #if no user selection, a filter selection is forced (enlarging the canvas)
        deltaplus = int(float((newwidth+newheight)*0.040))
    else:
        deltaplus = max(1,int(float((newwidth+newheight)*0.001)))  # was 0.000
    #info_msg ("deltaplus = "+str(deltaplus))    
    newwidthE = newwidth+deltaplus*2
    newheightE = newheight+deltaplus*2
    pdb.gimp_image_resize (newImage, newwidthE, newheightE, deltaplus, deltaplus)
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------
    if noUsrSel == True:   #if no user selection, a filter selection is forced (the central part)
        #info_msg ("no existing selection, default created")
        selectionWidth=newwidth 
        selectionHeight=newheight 
        pdb.gimp_image_select_rectangle (newImage, CHANNEL_OP_REPLACE,
                                              deltaplus, deltaplus,
                                              selectionWidth, selectionHeight)
    else:   # avoid problem with full image selection (creating a selection smaller of 1 pixel)
        pdb.gimp_selection_shrink (newImage, 1)
    (NE, sx1, sy1, sx2, sy2) = pdb.gimp_selection_bounds (newImage)
    selectionWidth=sx2-sx1 
    selectionHeight=sy2-sy1
    scentreX = selectionWidth/2
    scentreY = selectionHeight/2
    #info_msg ("selection bounds="+str(sx1)+","  +str(sy1)+","+str(sx2)+","+str(sy2)) 
        
    selection = pdb.gimp_selection_save (newImage)
     
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------
    
    # copy and desaturate the drawable
    
    theDesaturatedLayer = newBG.copy()
    pdb.gimp_image_add_layer (newImage, theDesaturatedLayer, 0)
    pdb.gimp_desaturate_full (theDesaturatedLayer, 1)       
    pdb.gimp_item_set_name (theDesaturatedLayer, "Base Layer Desaturated")
    
    # copy and emboss
    
    theEmbossedLayer = pdb.gimp_layer_new_from_drawable (theDesaturatedLayer, newImage)
    pdb.gimp_image_add_layer (newImage, theEmbossedLayer, 0)
    pdb.gimp_item_set_name (theEmbossedLayer, "Base Layer Desaturated & Embossed")
    pdb.gimp_layer_add_alpha (theEmbossedLayer)
    theEmbossedLayer.mode = DARKEN_ONLY_MODE
    pdb.plug_in_emboss (newImage, theEmbossedLayer, 210.0, 110.0, 15, 1)
    pdb.plug_in_erode (newImage, theEmbossedLayer, 0, 0, 1.0, 15, 0, 255)
    
    # visible 1
    
    theVisibleLayer1 = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible 1")
    pdb.gimp_image_add_layer (newImage, theVisibleLayer1, 0)
    pdb.gimp_layer_add_alpha (theVisibleLayer1)
    pdb.gimp_brightness_contrast (theVisibleLayer1, 5, 5)
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

    # Restore Original Image on top
    
    theSourceLayer1 = newBG.copy()
    pdb.gimp_image_add_layer (newImage, theSourceLayer1, 0)
    theSourceLayer1.name = "Source image"
    theSourceLayer1.mode = COLOR_MODE
    pdb.gimp_layer_add_alpha (theSourceLayer1)
    pdb.gimp_floating_sel_anchor (pdb.gimp_edit_named_paste (theSourceLayer1, realBufferName, False))
    pdb.gimp_layer_set_mode (theSourceLayer1, COLOR_MODE)
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

    #   duplicate and set to hardlight
    theSourceLayer2 = posterizeIT (theSourceLayer1, 32)     # ancillary routine
    theSourceLayer2.name = "Source Image posterized (256)"
    pdb.gimp_image_add_layer (newImage, theSourceLayer2, 0)
    pdb.gimp_layer_add_alpha (theSourceLayer2)
    pdb.gimp_hue_saturation (theSourceLayer2, 0 , 0 , 0 , +75)
    pdb.gimp_layer_set_mode (theSourceLayer2, HARDLIGHT_MODE)
    pdb.gimp_layer_set_opacity (theSourceLayer2, 100.0)
    
    # visible 2
    
    theVisibleLayer2 = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible 2")
    pdb.gimp_image_add_layer (newImage, theVisibleLayer2, 0)
    pdb.gimp_layer_add_alpha (theVisibleLayer2)
    
    pdb.plug_in_autocrop_layer (newImage, theVisibleLayer2)
    visiblewidth = theVisibleLayer2.width
    visibleheight = theVisibleLayer2.height
    (offsX, offsY) = pdb.gimp_drawable_offsets (theVisibleLayer2)
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

    ###  Add the Pattern
    
    #  textures must be available/selectable
    
    if namesFilter <> "None" :          # choose among a set of patterns
        (LTEXT, LLUMA, LCOLR)=load_patterns(namesFilter)
        rn = random.randrange (0, len(LTEXT))
        chosenPattern = LTEXT[rn]       # choose randomly
        # find measn colours values
        (meanR,a,b,c,d,e) = pdb.gimp_histogram (theSourceLayer2, HISTOGRAM_RED, 0, 255)
        (meanG,a,b,c,d,e) = pdb.gimp_histogram (theSourceLayer2, HISTOGRAM_GREEN, 0, 255)
        (meanB,a,b,c,d,e) = pdb.gimp_histogram (theSourceLayer2, HISTOGRAM_BLUE, 0, 255)
        maxV = max(meanR,meanG,meanB)
        minV = min(meanR,meanG,meanB)
        # stretch them
        adjR=meanR; adjG=meanG; adjB=meanB
        if maxV == meanR:
            adjR == 240
        elif maxV == meanG:
            adjG == 240
        else:
            adjB == 240
        if minV == meanR:
            adjR == 16
        elif minV == meanG:
            adjG == 16
        else:
            adjB == 16
        if selMethod == 0 : # best fit in luma
            one_color = (meanR, meanG, meanB, 255)
            chosenPattern = choose_appropriate_texture_by_LUMA(one_color, LLUMA, LTEXT)
        elif selMethod == 1 : # best fit in colour
            one_color = (adjR, adjG, adjB, 255)
            chosenPattern = choose_appropriate_texture_by_COLOUR(one_color, LCOLR, LTEXT)
        elif selMethod == 2 : # best fit in colour&luma
            one_color = ((adjR+meanR)/2, (adjG+meanG)/2, (adjB+meanB)/2, 255)
            chosenPattern = choose_appropriate_texture_by_LUMACOLOUR(one_color, LCOLR, LLUMA, LTEXT)
        else :    # use random texture  
            rn = random.randrange (0, len(LTEXT))
            chosenPattern = LTEXT[rn]     
    else:    
        chosenPattern = singlePattern
    #info_msg ("chosen pattern="+chosenPattern)
    
    (patternWidth, patternHeight, bpp) = pdb.gimp_pattern_get_info (chosenPattern)
    thePatternLayer = pdb.gimp_layer_new (newImage, patternWidth, patternHeight,
                                                RGBA_IMAGE, "Choosen texture (1st copy)", 100.00, HARDLIGHT_MODE)
    pdb.gimp_image_add_layer (newImage, thePatternLayer, 0)
    pdb.gimp_context_set_pattern (chosenPattern)
    pdb.gimp_edit_fill (thePatternLayer, PATTERN_FILL)
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------
    #info_msg ("visiblewidth="+str(visiblewidth))
 
    if PatternAction == 0   : # rescale
        #pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, visiblewidth, visibleheight)
        pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, selectionWidth, selectionHeight)  # try with sw,sh
        if noUsrSel == False:
            pdb.gimp_layer_translate (thePatternLayer, sx1, sy1)
               
    if PatternAction == 1   : #  make seamless
        #SLP_name = seamlessIT (chosenPattern)      # use ancillary routine
        SLP_name = seamlessIT (chosenPattern, myslash)      # added for compatibility with Linux OS
        pdb.gimp_context_set_pattern (SLP_name)
        thePatternLayer = pdb.gimp_image_get_active_layer (newImage)
        #pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, visiblewidth, visibleheight)
        pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, selectionWidth, selectionHeight)  # try with sw,sh
        if noUsrSel == False:
            pdb.gimp_layer_translate (thePatternLayer, sx1, sy1)
        #pdb.gimp_drawable_fill (thePatternLayer, PATTERN_FILL)
        pdb.gimp_edit_fill (thePatternLayer, PATTERN_FILL)  # try using edit     <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 
    if PatternAction == 2   : # make seamless and rescale
        SLP_name = seamlessIT (chosenPattern, myslash)      # use ancillary routine
        pdb.gimp_context_set_pattern (SLP_name)
        thePatternLayer = pdb.gimp_image_get_active_layer (newImage)
        #pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, visiblewidth, visibleheight)
        pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, selectionWidth, selectionHeight)  # try with sw,sh
        if noUsrSel == False:
            pdb.gimp_layer_translate (thePatternLayer, sx1, sy1)
               
    if PatternAction == 3   : # do nothing on the pattern, hence...refill
        #pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, visiblewidth, visibleheight)
        pdb.gimp_item_transform_scale (thePatternLayer, 0, 0, selectionWidth, selectionHeight)  # try with sw,sh
        if noUsrSel == False:
            pdb.gimp_layer_translate (thePatternLayer, sx1, sy1)
        #pdb.gimp_drawable_fill (thePatternLayer, PATTERN_FILL)
        pdb.gimp_edit_fill (thePatternLayer, PATTERN_FILL)  # try using edit     <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        
    pdb.gimp_layer_set_mode (thePatternLayer, NORMAL_MODE)          # was HARDLIGHT
    pdb.gimp_layer_set_opacity (thePatternLayer, 50.00)             # was 100.00
    
    pdb.gimp_layer_translate (thePatternLayer, offsX, offsY)
    pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)
    pdb.gimp_selection_invert (newImage)
    pdb.gimp_edit_cut (thePatternLayer)
    pdb.gimp_selection_invert (newImage)
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

    # B&W over, then color white to transparent (thus replaced by below pattern colors)
    
    theBWout2 = pdb.gimp_layer_new_from_drawable (theVisibleLayer2, newImage)
    pdb.gimp_image_add_layer (newImage, theBWout2, 0)
    pdb.gimp_layer_add_alpha (theBWout2)
    pdb.gimp_desaturate (theBWout2)         # < ============================
    theBWout2.name = "BW out without white"
    theBWout2.opacity = 75.00   # was 100
    theBWout2.mode = NORMAL_MODE
    pdb.plug_in_colortoalpha (newImage, theBWout2, ((255,255,255,255)))     
        
    # Duplicate the PatternLayer to increase colors and texture
        
    thePatternLayer2 = pdb.gimp_layer_new_from_drawable (thePatternLayer, newImage)
    pdb.gimp_image_add_layer (newImage, thePatternLayer2, 0)
    pdb.gimp_layer_add_alpha (thePatternLayer2)
    thePatternLayer2.name = "Choosen texture (2nd copy)"
    thePatternLayer2.opacity = 65.00
    thePatternLayer2.mode = MULTIPLY_MODE   #HARDLIGHT_MODE                           
    pdb.gimp_hue_saturation (thePatternLayer2, 0, 0, 0, -30)       
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------
    
    if noUsrSel == False:    # if called with a selection create immediately the under-layer
        theUnderLayer = pdb.gimp_layer_new (newImage, newwidthE, newheightE,
                                                RGBA_IMAGE, "Under Layer", 100.00, MULTIPLY_MODE)
        pdb.gimp_image_add_layer (newImage, theUnderLayer, 0)
        # ============= attempt using gradient ==========================================================================
        pdb.gimp_selection_none (newImage)
        pdb.gimp_context_set_gradient ("Shadows 3")
        # attention! fn "edit_blend" does not work on trasparency!!
        pdb.gimp_edit_fill (theUnderLayer, WHITE_FILL)
        pdb.gimp_edit_blend (theUnderLayer, CUSTOM_MODE, NORMAL_MODE, GRADIENT_CONICAL_SYMMETRIC, 100.0, 0, REPEAT_NONE,
                            #reverse, supers, depth, thresh, dither,
                             False,   False,  1,     0,      0,
                            #newwidth*0.55, newheight*0.55, newwidth*1.1, newheight*0.55)
                            scentreX, scentreY, newwidth*1.1, newheight*0.55)
        pdb.gimp_brightness_contrast (theUnderLayer, -100, +25)     # was -70 +35
        bluramt = (newwidthE+newheightE)/200
        pdb.plug_in_gauss_iir2 (newImage, theUnderLayer, bluramt, bluramt)
        theUnderLayer.mode = MULTIPLY_MODE
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)    # the full selection
        # try reducing the cut
        selWH = selectionWidth+selectionHeight
        border = int(selWH/120)     # was 100
        pdb.gimp_selection_shrink (newImage, border/4)
        # end addition
        pdb.gimp_edit_cut (theUnderLayer)
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)    # restore the full selection
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

        # ===============================================================================================================

    # visible 3
    
    theVisibleLayer3 = pdb.gimp_layer_new_from_visible (newImage, newImage, "Visible 3")
    pdb.gimp_image_add_layer (newImage, theVisibleLayer3, 0)
    #pdb.gimp_layer_add_alpha (theVisibleLayer3)        # commented out
    #pdb.gimp_display_new (newImage)
    #return()

    selWH = selectionWidth+selectionHeight
    border = int(selWH/100)      
    # ---------------------------------------------------------------------------
    # adjust on the base of the path length relative to the raw border
    #info_msg ("border raw = "+str(border))
    pdb.plug_in_sel2path (newImage, theVisibleLayer3)
    activevectors = pdb.gimp_image_get_active_vectors (newImage)
    totallength = 0
    (nrs, ids) = pdb.gimp_vectors_get_strokes (activevectors)
    for stroke in ids:
        length = pdb.gimp_vectors_stroke_get_length (activevectors, stroke, 0.01)
        totallength = totallength+length
    #info_msg ("total selection contour length = "+str(totallength))
    selcontour = selWH*2
    #info_msg ("selection raw square contour length = "+str(selcontour))
    scfactor = totallength/selcontour
    #info_msg ("selection contour factor = "+str(scfactor))
    border = int(border/scfactor)
    #info_msg ("border adjusted = "+str(border))
    
    # ---------------------------------------------------------------------------
    azimuth_value = 110.0
    
    if doContour == True:       # simply create a contour and fill with colour
        pdb.gimp_context_set_foreground ((180,150,25,255))
        pdb.gimp_context_set_feather (False)
        pdb.gimp_context_set_feather_radius (0.00, 0.00)
        pdb.gimp_context_set_opacity (100.0)
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)
        pdb.gimp_selection_shrink (newImage, int(border/2))
        pdb.gimp_selection_border (newImage, border)                                 # the border
        pdb.gimp_edit_clear (theVisibleLayer3)
        pdb.gimp_edit_fill (theVisibleLayer3, FOREGROUND_FILL)
        pdb.gimp_edit_fill (theVisibleLayer3, FOREGROUND_FILL)                       # twice seems better
        
    elif doChisel == True:                      # we may use chisel.scm
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)         # the full selection
        if noUsrSel == True:    # if called without a selection
            border = border*1.5
            width_value = border
            pdb.gimp_selection_grow (newImage, border*2/3)                               # <=======   added *2/3
            pdb.gimp_selection_border (newImage, width_value/2)
            pdb.gimp_context_set_foreground ((127,32,32,255))
        else:
            #border = border*0.50
            width_value = border
        if noUsrSel == True:    # if called without a selection
            pdb.gimp_edit_fill (theVisibleLayer3, FOREGROUND_FILL)                       # chisel does not work on transaprency
            pdb.gimp_edit_fill (theVisibleLayer3, FOREGROUND_FILL)                       # chisel does not work on transaprency
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)         # the full selection
        #pdb.gimp_selection_grow (newImage, width_value)
        ###pdb.gimp_selection_grow (newImage, border*1/4)                               # test
        if noUsrSel == False:    # if called with a selection                        # new addition   
            filter_mode = 0      # chisel/carve
            pdb.gimp_selection_grow (newImage, border*1/8)                               # test
        else:    
            filter_mode = 1      # chisel/carve
            pdb.gimp_selection_grow (newImage, border*1/4)                               # test
        pdb.gimp_context_set_foreground ((32,32,32,255))        # changed from 0.0.0
        pdb.gimp_context_set_background ((223,223,223,223))     # added
                                           # was 120
        elev_value = border*1.5 # adaptive
        depth_value = border    # adaptive
        ccRoundness = 0.00      # can't be zero? was 0.01
        #info_msg ("chisel parms (width,elev,depth)="+str(width_value)+","+str(elev_value)+","+str(depth_value))
        #pdb.gimp_display_new (newImage)
        #return()       
        pdb.script_fu_chisel(newImage, theVisibleLayer3,
                             width_value,             #bevel width
                             3,         #bevel softness,        was 10*
                             0.0,             #bevel roundness (0.1-1.0),
                             0.0,             #bevel power 
                             azimuth_value,           #azimuth  
                             elev_value,              #elevation
                             depth_value,             #depth,  
                             filter_mode,             #mode: 0=chisel,1=carve,
                             0,                       #location,
                             0.0,                     #post-effect blur,
                             False)                   #keep)
        #pdb.gimp_display_new (newImage)
        #return()       
        # new layer created
        LA = newImage.layers
        bevel = bevel=LA[0]
        theVisibleLayer3 = pdb.gimp_image_merge_down (newImage, bevel, 2)
        #pdb.gimp_display_new (newImage)
        #return()
        if noUsrSel == True:    # if called without a selection (added)
        ####twice to better contrast
            pdb.script_fu_chisel(newImage, theVisibleLayer3,
                             width_value,             #bevel width
                             3,         #bevel softness,
                             0.0,             #bevel roundness (0.1-1.0),
                             0.0,             #bevel power, #test
                             azimuth_value,           #azimuth  
                             elev_value,              #elevation
                             depth_value,             #depth,  
                             filter_mode,             #mode: 0=chisel,1=carve,
                             0,                       #location,
                             0.0,                     #post-effect blur,
                             False)                   #keep)
            #pdb.gimp_display_new (newImage)
            #return()       
            # new layer created
            LA = newImage.layers
            bevel = bevel=LA[0]
            theVisibleLayer3 = pdb.gimp_image_merge_down (newImage, bevel, 2)
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------
    else:
        #info_msg ("hard-bump being implemented")
        tt_layer = theVisibleLayer3.copy()
        pdb.gimp_image_add_layer (newImage, tt_layer, 0)
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)
        pdb.gimp_selection_invert (newImage)
        pdb.gimp_edit_cut (tt_layer)
        #pdb.gimp_display_new (newImage)
        #return()
        
        #saves to channel to be used as bump map later
        pdb.gimp_selection_feather(newImage,15)
        channel = pdb.gimp_selection_save(newImage)
        
        # fix shadow length
        shadowlength = int(selWH/50)
        #info_msg ("shadow length = "+str(shadowlength))
        
        #bump map
        pdb.gimp_selection_none(newImage)
        pdb.plug_in_bump_map(newImage,tt_layer,channel,azimuth_value,27,65,0,0,0,50,True,True,0)
        
        position = pdb.gimp_image_get_item_position(newImage,tt_layer)
        shadow1 = pdb.gimp_layer_new_from_drawable(tt_layer,newImage)
        pdb.gimp_image_insert_layer(newImage,shadow1,None,position+1)
        pdb.gimp_layer_set_lock_alpha(shadow1,True)
        pdb.gimp_context_set_foreground((0,0,0))
        pdb.gimp_edit_fill(shadow1,FOREGROUND_FILL)
        pdb.gimp_layer_set_lock_alpha(shadow1,False)
        pdb.gimp_item_set_name(shadow1,"Shadow 1")
        pdb.plug_in_mblur(newImage,shadow1,0,shadowlength,(-azimuth_value) % 360,0,0)
        
        position = pdb.gimp_image_get_item_position(newImage,shadow1)
        shadow2 = pdb.gimp_layer_new_from_drawable(shadow1,newImage)
        pdb.gimp_image_insert_layer(newImage,shadow2,None,position+1)
        pdb.gimp_item_set_name(shadow2,"Shadow 2")
        
        #color balance
        pdb.gimp_color_balance(tt_layer,1,True,10,0,-10)
        pdb.gimp_color_balance(tt_layer,2,True,5,-15,-30)
        
        #pdb.gimp_display_new (newImage)
        #return()
        
        #merge layers
        LA = newImage.layers
        theVisibleLayer3 = pdb.gimp_image_merge_down (newImage, LA[2], 0)
        theVisibleLayer3 = pdb.gimp_image_merge_down (newImage, LA[1], 0)
        theVisibleLayer3 = pdb.gimp_image_merge_down (newImage, LA[0], 0)

    # =====================================================================
    
    pdb.gimp_selection_none (newImage)
    
    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

    # visibility on for all layers  
    LA = newImage.layers
    for alayer in LA:
        pdb.gimp_layer_set_visible (alayer, True)

    if noUsrSel == True:    # if called without a selection create now the under-layer
        factor = 0.000      # no enlargement
        shift = int((newwidthE+newheightE)*factor)
        delta = shift*2
        pdb.gimp_image_resize (newImage, newwidthE+delta, newheightE+delta, shift, shift)
        theUnderLayer = pdb.gimp_layer_new (newImage, newwidthE+delta, newheightE+delta,
                                                RGBA_IMAGE, "Under Layer", 100.00, NORMAL_MODE)
        pdb.gimp_image_add_layer (newImage, theUnderLayer, 1)
        pdb.gimp_context_set_pattern ("Bricks")          
        pdb.gimp_edit_fill(theUnderLayer, PATTERN_FILL)
        # ------------ increase the chosen pattern for BG ----------------------------------
        bgsize = widthheight/20
        tempImage = pdb.gimp_image_new(bgsize, bgsize, RGB)
        tempLayer = pdb.gimp_layer_new(tempImage,96,96,RGB_IMAGE,"temp",100.0,NORMAL_MODE)
        tempImage.add_layer(tempLayer, 0)
        pdb.gimp_drawable_fill (tempLayer, PATTERN_FILL)
        pdb.gimp_item_transform_scale (tempLayer, 0, 0, bgsize, bgsize)  
        d_dir = pdb.gimp_gimprc_query ("gimp_dir")
        T_name = "temp_pat"
        #T_name2 = d_dir+"\\patterns\\"+T_name+".pat"
        T_name2 = d_dir+myslash+"patterns"+myslash+T_name+".pat"        # changed for compatibility with Linux OS
        pdb.file_pat_save (tempImage, tempLayer, T_name2, "", T_name)
        pdb.gimp_patterns_refresh()
        pdb.gimp_context_set_pattern (T_name)          
        pdb.gimp_edit_fill(theUnderLayer, PATTERN_FILL)
        # -----------------------------------------------------------------------------------
        pdb.gimp_brightness_contrast (theUnderLayer, -50, +50)
        pdb.gimp_colorize (theUnderLayer, 24, 70, 28)   # colour for bricks
        theUnderLayer.mode = NORMAL_MODE
        
        pdb.gimp_image_select_item (newImage, CHANNEL_OP_REPLACE, selection)         # the full selection
        border = int(selWH/120) 
        pdb.gimp_selection_grow (newImage, border*2/3)                               # <=======   added *2/3
        pdb.gimp_selection_border (newImage, border/2)
        pdb.gimp_context_set_foreground ((127,32,32,255))
        pdb.gimp_edit_fill (theUnderLayer, FOREGROUND_FILL)
        pdb.gimp_edit_fill (theUnderLayer, FOREGROUND_FILL)
        pdb.gimp_selection_none (newImage)

    #----------------------------------------------------------------------
    #pdb.gimp_display_new (newImage)
    #return()
    #----------------------------------------------------------------------

    # last step: flatten the image if requested by the user (default=yes)
    
    if (doFlatten == True) :
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theDesaturatedLayer, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theEmbossedLayer, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theVisibleLayer1, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theSourceLayer1, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theSourceLayer2, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theVisibleLayer2, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, thePatternLayer, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theBWout2, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, thePatternLayer2, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theUnderLayer, 0)
        flattenedLayer = pdb.gimp_image_merge_down (newImage, theVisibleLayer3, 0)
        pdb.gimp_item_set_name(flattenedLayer,
                               st+" "+sourcename+" (pattern="+chosenPattern+")")
    
    if origtype == GRAY :
        pdb.gimp_image_convert_grayscale (inImage)   # restore
    
    pdb.gimp_image_undo_group_end(inImage)                                                                
    pdb.gimp_displays_flush()
    pdb.gimp_display_new(newImage)
    
    pdb.gimp_context_set_foreground(currFG)
    pdb.gimp_context_set_background(currBG)
    pdb.gimp_context_pop

    if (doFlatten == True) :
        return(flattenedLayer)
    else :
        return()

# This is the plugin registration function  
register(
    "PrintOnStones2",
    "creates an image looking as a print on stone (v.2)",
    "creates an image looking as a print on stone (v.2)",
    "Diego", 
    "Diego Nassetti ", 
    "May 2017",
    "PrintOnStones2",
    "RGB*,GRAY*",
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "drawable", "Input drawable", None),
      (PF_ADJUSTMENT, "size", "Width+Height of New Image"+"\n (0=same as input)", 3000, (0, 8000, 200)),
      (PF_STRING, "filter", "Filter for patterns"+"\n (group of stone patterns)", "None"),
      (PF_PATTERN, "singlepattern", "or Single Pattern"+
                                    "\n >>>> Warning! <<<<"+"\n used ONLY if Filter set to None", "Lightning"),         
      (PF_OPTION, "action", "Action on Pattern", 0, ("Rescale", "Make-seamless",
                                                     "Make-seamless and Rescale", "None")),
      (PF_OPTION, "selmethod", "Patterns selection method", 3, ("Best Luma",
                                                                "Best Colour",
                                                                "Best Luma&Colour",
                                                                "Random",
                                                               )),
      (PF_OPTION, "contour", "contour method", 1, ("Simple contour",
                                                    "Use (extrn) chisel script"+"\n (you must have it installed and active)",
                                                    "Use TT hardbevel (embedded)",
                                                  )),
      (PF_TOGGLE, "flatten", "Flatten Image?", True),
    ],
    [
      (PF_DRAWABLE, "odrawable", "Output drawable", None),  
    ],
    PrintOnStones2,
    menu="<Image>/Diego/Artistic",
    )

main()
   